------------------------------------------------------------------------
--  Object:  CSE 561 HW 10
--  Title :  Bouncer with any width
--  Author:  Chung-Cheng Kuo  (SSN: 120-82-4266)
--  Date  :  11/10/1994
------------------------------------------------------------------------
--  Introduction:
--    1. The bouncer has a internal reset signal to set the first register
--       '1' after beginning.  It is descibed as the following:
--       (1) It is reset to '1' at time=0.
--       (2) It is changed to '0' after the first clock comes in.
--

-- use work.all;

entity DFF is
   port ( D,CLK : bit;
          Q : inout bit);
end DFF;

architecture beh of DFF is
begin 
   q <= d when (clk'event and clk='1') else 
        q; 
end beh;


entity mplx2 is
   port ( data0, data1, control: bit;
          dataout : out bit);
end mplx2;

architecture beh of mplx2 is 
begin
   dataout <= data0 when control='0' else
              data1;
end;


entity bouncer is
   generic ( n: positive); -- size
   port ( clk : bit;
          reg : out bit_vector(0 to n-1) );
end bouncer;

architecture struc of bouncer is
   component dff port ( d, clk: bit; q : inout bit); end component;   
   component mplx2 port ( data0, data1, control: bit; dataout : out bit); 
                                                               end component;
   signal control, zero : bit;
   signal reset : bit := '1';
   signal d, q : bit_vector(0 to n-1);
begin
   control <= '0' when q(0)='1' and q(n-1)='0' else    -- shift right
              '1' when q(0)='0' and q(n-1)='1' else    -- shift left
              control;			           -- keep current direction
   reset <= '0' after 2 ns when clk'event and clk='1' else reset;

   -- Generate general components
   L1: for i in 1 to n-2 generate
          L11: mplx2 port map (q(i-1),q(i+1),control,d(i));
          L12: dff port map (d(i), clk, q(i));
          L13: reg(i) <= q(i);
       end generate;

   --- Generate the first component
   L21: mplx2 port map (reset,q(1),control,d(0));
   L22: dff port map(d(0),clk,q(0));
   L23: reg(0) <= q(0);

   --- Generate the last component
   L31: mplx2 port map (q(n-2),zero,control,d(n-1));
   L32: dff port map(d(n-1),clk,q(n-1));
   L33: reg(n-1) <= q(n-1);

end struc;



entity test_bouncer is end;
architecture struct of test_bouncer is
   constant n : integer := 6;
   component bouncer 
       generic (n : positive);
       port (clk : bit; reg : out bit_vector(0 to n-1)); 
   end component;

   signal clock  : bit;
   signal lights : bit_vector (0 to n-1);
begin
   clock <= not clock after 5 ns;
   L: bouncer generic map(n) port map(clock, lights);
end struct;


