--Homework Assignment 12  
--CSE561 Fall 1994
 
--This problem is related to the orderly sharing of communication
-- privileges on a shared bus.
-- You are to design a shared bus interface system, and
--simulate it with any tests you like. We will also simulate it
--in the process of grading it.
-- 
--Specification: There is a 4-bit data bus that is shared. D
--     There is a 2-bit address that is shared.  A
--     There is are two 1-bit command busses shared, R, W
--     There is a shared 1-bit acknowledge signal shared,  READY
--     It is essential that only one source be
--     controlling these four busses. A wired-or resolution
--     function models the interaction on the shared busses.
--        
--     Modules will be connected to the buss that
--     will need to take control from time to time.
--     There are some bus signals to support sharing the bus.
--     There will be a shared 1-bit BUSY bus.
--     There will be a shared 1-bit REQUEST bus.
--     Each module will have an ARBITRATION-In, and
--     an ARBITRATION-OUT ports connected up
--     in daisy chain style.
--     Because of the daisy chain connection signals will
--      be delayed in the daisy chain more than on other
--      bus signals. 
--     Here are the communication rules for a master-slave
--     Read transaction:
--     Master                    Slave
--     Puts address on A
--     Asserts R
--                               The addressed slave
--                              notices A and R so it
--                              puts data on D and asserts
--                              READY
--     Notices READY so reads
--     data from D and unasserts
--     R 
--                               Notices R = 0 so unasserts 
--                               READY.
--    
--     THE TRANSACTION IS OVER
--
--     Write transaction:
--     Master                    Slave
--     Puts address on A
--     Data on D
--     Asserts W
--                               The addressed slave
--                              notices A and W so it
--                              reads data on D and asserts
--                              READY
--     Notices READY so removes
--     data from D and unasserts
--     W 
--                               Notices W = 0 so unasserts 
--                               READY.
-- 
--     THE TRANSACTION IS OVER
--
--     Those are the communication rules when the bus has a
--     master. We need rules to be followed for a device on the 
--     bus to become master and to stop being the master.
--
--     The rules are:
--     If a module needs the bus, it waits until REQUEST = '0' 
--     indicating nor REQUESTS and then it asserts REQUEST. Of course,
--     if REQUEST = '0 when a unit wants to become master it just
--     asserts the REQUEST, 
--     When it asserts REQUEST, it sets ARBOUT = '0' to make sure 
--     ARBIN = '0' for all downstream units.
--     Then it waits until the bus is not BUSY, and ARB-IN = 1
--     Then the module asserts BUSY. 
--     At this point the module can make use of the
--     shared bus for communication. The module, which is now master,
--     removes its assertion of REQUEST, and it
--     sets ARBOUT = ARBIN.
--     When the module is ready to give up the bus  
--     it removes it assertion of BUSY.
--     Since the master module can unassert REQUEST, and
--     set ARBOUT=ARBIN prior to unasserting BUSY.
--     the time required for arbitration overlaps
--     the communication time, hopefully reducing the
--     time delay associated with arbitration.
--  
--     Nothing so far has required any design on your part.
--     You are to design a component that will be instantiated
--     a number of times and attached to the bus.
--     Call it a BOX. It has ports.  The only ports is has
--     are the ports described for the bus above.
--     Each BOX has three generic parameters that are set
--     at instantiation time: 1. name: bit_vector -- BOX id
--                            2. delay: time; -- the delay of ARBOUT
--                            3. interval: time; -- related to this problem
--     Each BOX behaves essentially the same. 
--     As a slave
--     it monitors the bus and when it sees it own name
--     as address bits it engages in the required transaction.
--     If the transaction is a write, it reads the data line
--     and saves the result in a register called SAVE. 
--     If the transaction is a read, the BOX should put its SAVE
--     value on the data-bus.
--     This is the slave operation of BOX responding to the requests of 
--     another BOX which is bus master.
--     In addition, each BOX has a time-out interval. At time intervals 
--     specified by
--     the generic parameter, interval,  each BOX tries to take possession
--     of the bus. When it becomes master, it reads from the BOX
--     with a name one less than its own, and adds the value
--     read to SAVE, saving the result in SAVE. Then it writes
--     that value to the BOX with address one more than its own.
--     Then it is done with the bus until another interval of
--     delay. In addition, for each BOX the initial value in SAVE should
--     be its own name.
--     Since we are modelling an environment with some assumed delay
--     all the assignments you make to ARBOUT should be of the form:
--     ARBOUT <= <value> after delay; -- delay is a generic time
--     For example, if your control system generates a signal called
--     STOP_DAISY whenever the daisy should be stopped you could write:
--     ARBOUT <= '0'   after delay  when STOP_DAISY else
--               ARBIN after delay;
--     In addition, you should make all transfers to the bus have some 
--     delay. I suggest 100 ns as this delay. The material below uses this
--     value.  That defines BOX.
--      
--     There is an entity called SYSTEM that has no
--     ports, but which has 4 BOXes connected up appropriately.
--     A testbench is really not needed since SYSTEM has no
--     ports and since the BOXes will become active on their 
--     own.
--

--  The material below is intended to help you. If you do not find it helpful
--  change it any way you like, or start from scratch. The main goal is to 
--  meet the specifications.
package resolv is
function wired_or (Input: bit_vector) return bit;
subtype r_bit is wired_or bit;
type r_bit_vector is array(integer range<>) of r_bit;
end resolv;

package body resolv is

function wired_or (Input: bit_vector) return bit is
variable result : bit := '0';
begin
  for i in input'range loop
    if (input(i) = '1') then
       result := '1'; 
       exit;
    end if;
    end loop;
    return result;
end wired_or;
end resolv;

use work.resolv.all;

entity BOX is 
   generic( id : r_bit_vector(0 to 1);   --  BOX id
            delay : time;               -- delay of ARBOUT
            interval: time );           -- activity interval
   port( A: inout r_bit_vector(0 to 1);
         D: inout r_bit_vector(0 to 3); 
         R: inout r_bit;
         W: inout r_bit;
         READY : inout r_bit;
         BUSY: inout r_bit;
         REQUEST: inout r_bit;
         ARBIN: in bit;
         ARBOUT: out bit);
end BOX;

architecture beh of BOX is
signal SAVE : r_bit_vector(0 to 3);
signal DEC, INC : r_bit_vector(0 to 1) ; -- address of neighbors
signal STOP_DAISY, add2save, loadsave1, loadsave2 : boolean; 
signal master_read, master_write, read_respond, write_respond : boolean ;
signal write_complete, read_complete : boolean;
constant bus_delay : time := 100 ns;
begin

process -- initialization of DEC, INC, this BOXes neighbors
begin DEC  <= not(id(0) xor id(1)) & (not id(1));
      INC  <= (id(0) xor id(1)) & (not id(1));
wait; 
end process;

process-- control of SAVE
variable c: bit;
begin
  if (now = 0 ns) then SAVE <= "00" & id; end if;
  if  add2save  then
     c := '0';
     for i in 3 downto 0 loop
        SAVE(i) <= SAVE(i) xor D(i) xor c;
        c := (c and (SAVE(i) or D(i))) or (D(i) and SAVE(i));
     end loop;
   end if;
   if (loadsave1 or loadsave2) then SAVE <= D; end if;
   wait on loadsave1, loadsave2, add2save;
end process;  
    

process -- the master does a read 
  begin
    wait on master_read;        -- a pulse from the control unit
    A <= DEC after bus_delay;
    R <= '1' after bus_delay;
    wait until (READY = '1');
    add2save <= true, false after 1 ns;
    R <= '0' after bus_delay;
    A <= "00" after bus_delay;
    wait until (READY = '0');
    read_complete <= true,false after 1 ns;
end process;


process -- the master does a write
  begin
    wait on master_write;       -- a pulse from the control unit  
    A <= INC after bus_delay;
    W <= '1' after bus_delay;
    D <= SAVE after bus_delay;
    wait until (READY = '1');
    W <= '0' after bus_delay;
    A <= "00" after bus_delay;
    D <= "0000" after bus_delay;
    wait until (READY = '0');
    write_complete <= true, false after 1 ns;
end process;

ARBOUT <= '0'    after delay when STOP_DAISY else  -- STOP_DAISY comes from
           ARBIN after delay;                       -- control

-- SLAVE behavior
-- You are to write the VHDL for how the BOX behaves as a slave
  
-- MASTER behavior control
process 
  begin
    wait for interval;
-- you will put VHDL code in here
-- it should acquire and release the bus in an orderly way.
-- once it has the bus it should cause the necessary read and write
-- and release the bus
-- try to have the bus arbitration time overlap the communication time
end process;

end beh;

use work.resolv.all;
use work.all;
 
entity system is end;
-- no ports

architecture beh of system is

component BOX  
   generic( id : r_bit_vector(0 to 1);
            delay : time;
            interval : time);
   port( A: inout r_bit_vector(0 to 1);
         D: inout r_bit_vector(0 to 3); 
         R: inout r_bit;
         W: inout r_bit;
         READY : inout r_bit;
         BUSY: inout r_bit;
         REQUEST: inout r_bit;
         ARBIN: in bit;
         ARBOUT: out bit);
end component ;

signal A: r_bit_vector(0 to 1);
signal D: r_bit_vector(0 to 3); 
signal ARB: bit_vector(0 to 3);
signal R, W, READY, BUSY, REQ : r_bit;

begin
-- in this case, a complicated case, each BOX tries to get the bus
-- 1 us after it finishes.
-- Try simpler cases first, so you can easily predict expected behavior

        L0: box
             generic map( "00", 1 us, 1 us)
             port map( A, D, R,W,READY,BUSY,REQ, REQ, ARB(0));

        L1: box
             generic map( "01", 1 us, 1 us)
             port map( A, D, R,W,READY,BUSY,REQ, ARB(0), ARB(1));

        L2: box
             generic map( "10", 1 us, 1 us)
             port map( A, D, R,W,READY,BUSY,REQ, ARB(1), ARB(2));

        L3: box
             generic map( "11", 1 us, 1 us)
             port map( A, D, R,W,READY,BUSY,REQ, ARB(2), ARB(3));

end beh ;


