------------------------------------------------------------------------
--  Object:  CSE 561  Term Project 2
--  Title :  Pipelined Comparator
--  Author:  Joncheng Kuo  (SSN: 120-82-4266)
--  Date  :  12/  /1994
------------------------------------------------------------------------
-- 7-stage pipeline
-- with timing analysis in gate and register level
-- scalable pipeline stages
-- try to insert latches into result_unit

------------------------------------------------------------
--  inverter with mvl4
------------------------------------------------------------
use work.system4.all;
entity invert4 is
    generic (Tmin, Tmax : time);
    port (inval :  in mvl4;
	  outval : inout mvl4);
end invert4;

architecture beh of invert4 is
begin
   process(inval)
     variable temp : mvl4;
     begin
	  temp := not inval;
	  if (outval /= temp)
	     then  outval <= 'X' after Tmin, temp after Tmax;
	     else  outval <= outval;
	   end if;
    end process;
end beh;


------------------------------------------------------------
--  nand2 with mvl4
------------------------------------------------------------
use work.system4.all;
entity nand2_4 is
    generic (Tmin, Tmax : time);
    port (inval1, inval2 :  in mvl4;
	  outval : inout mvl4);
end nand2_4;

architecture beh of nand2_4 is
begin
   process(inval1,inval2)
     variable temp : mvl4;
     begin 
          temp := not (inval1 and inval2); 
          if (outval /= temp)
             then  outval <= 'X' after Tmin, temp after Tmax;
             else  outval <= outval;
           end if;
     end process;
end beh;


------------------------------------------------------------
--  nand3 with mvl4
------------------------------------------------------------
use work.system4.all;
entity nand3_4 is
    generic (Tmin, Tmax : time);
    port (inval1, inval2, inval3 :  in mvl4;
	  outval : inout mvl4);
end nand3_4;

architecture beh of nand3_4 is
begin
   process(inval1,inval2,inval3)
     variable temp : mvl4;
     begin
	  temp := not (inval1 and inval2 and inval3);
	  if (outval /= temp)
	     then  outval <= 'X' after Tmin, temp after Tmax;
	     else  outval <= outval;
	   end if;
     end process;
end beh;


------------------------------------------------------------
--  and2 with mvl4
------------------------------------------------------------
use work.system4.all;
entity and2_4 is
    generic (Tmin, Tmax : time);
    port (inval1, inval2 :  in mvl4;
	  outval : inout mvl4);
end and2_4;

architecture beh of and2_4 is
begin
   process(inval1,inval2)
     variable temp : mvl4;
     begin
	  temp := inval1 and inval2;
	  if (outval /= temp)
	     then  outval <= 'X' after Tmin, temp after Tmax;
	     else  outval <= outval;
	   end if;
     end process;
end beh;


------------------------------------------------------------
--  nor2 with mvl4
------------------------------------------------------------
use work.system4.all;
entity nor2_4 is
    generic (Tmin, Tmax : time);
    port (inval1, inval2 :  in mvl4;
	  outval : inout mvl4);
end nor2_4;

architecture beh of nor2_4 is
begin
   process(inval1,inval2)
     variable temp : mvl4;
     begin
	  temp := not (inval1 or inval2);
	  if (outval /= temp)
	     then  outval <= 'X' after Tmin, temp after Tmax;
	     else  outval <= outval;
	   end if;
     end process;
end beh;


------------------------------------------------------------
--  or2 with mvl4
------------------------------------------------------------
use work.system4.all;
entity or2_4 is
    generic (Tmin, Tmax : time);
    port (inval1, inval2 :  in mvl4;
	  outval : inout mvl4);
end or2_4;

architecture beh of or2_4 is
begin
   process(inval1,inval2)
     variable temp : mvl4;
     begin
	  temp := inval1 or inval2;
	  if (outval /= temp)
	     then  outval <= 'X' after Tmin, temp after Tmax;
	     else  outval <= outval;
	   end if;
     end process;
end beh;


------------------------------------------------------------
--  xor2 with mvl4
------------------------------------------------------------
use work.system4.all;
entity xor2_4 is
    generic (Tmin, Tmax : time);
    port (inval1, inval2 :  in mvl4;
	  outval : inout mvl4);
end xor2_4;

architecture beh of xor2_4 is
begin
   process(inval1,inval2)
     variable temp : mvl4;
     begin
	  temp := inval1 xor inval2;
	  if (outval /= temp)
	     then  outval <= 'X' after Tmin, temp after Tmax;
	     else  outval <= outval;
	   end if;
     end process;
end beh;


------------------------------------------------------------
--  D Flip-Flop
------------------------------------------------------------
use work.system4.all;
entity DFF is
   generic (Tmin, Tmax : time);
   port ( D, clk : in mvl4;
	  Q : inout mvl4);
end DFF;

architecture beh of DFF is
begin
   process (clk)
   begin
      if clk'event and clk='1' then
	 if Q/=D then
	    Q <= 'X' after Tmin, D after Tmax;
	 else
	    Q <= Q;
	 end if;
      end if;
   end process;
end beh;


------------------------------------------------------------
--  n-bit latch
------------------------------------------------------------
use work.system4.all;
entity latch is
   generic (n : positive;          -- size
	    Tmin, Tmax : time);    -- gate delay time
   port ( cin : in mvl4_vector(n-1 downto 0);
	  clk : in mvl4;
	  cout : inout mvl4_vector(n-1 downto 0));
end latch;

architecture struc of latch is
   component DFF
      generic (Tmin, Tmax : time);
      port ( D, clk : in mvl4; Q : inout mvl4);
   end component;
begin
   L1: for i in 0 to n-1 generate
   L2:    DFF generic map (Tmin,Tmax) port map (cin(i),clk,cout(i));
       end generate;
end struc;



------------------------------------------------------------
--  full_adder
------------------------------------------------------------
use work.system4.all;
entity full_adder is
   generic (Tmin, Tmax : time);    -- gate delay time
   port ( a, b, cin : mvl4;
	  sum, cout : inout mvl4);
end full_adder;

-- behavier of FA
----------------------------------------
-- sum = a xor b xor c
-- carry = a.b + a.c + b.c
--       = a.(b+c) + b.c
--       = ~( ~(a.b) . ~(a.c) . ~(b.c) )
----------------------------------------
architecture beh of full_adder is
   constant stages : integer := 3;    -- number of gate stages in FA
begin
   process (a,b,cin)
      variable d_min, d_max : time;
   begin
      d_min := stages * Tmin;
      d_max := stages * Tmax;
      sum  <= 'X' after d_min, a xor b xor cin after d_max;
      cout <= 'X' after d_min,
	      (a and b) or (a and cin) or (b and cin) after d_max;
   end process;
end beh;

-- FA with 3 gate stages (using nand3)
----------------------------------------
architecture struc1 of full_adder is
   component and2_4
      generic (Tmin, Tmax : time);
      port ( inval1, inval2 : in mvl4; outval : inout mvl4);
   end component;
   component or2_4
      generic (Tmin, Tmax : time);
      port ( inval1, inval2 : in mvl4; outval : inout mvl4);
   end component;
   component nor2_4
      generic (Tmin, Tmax : time);
      port ( inval1, inval2 : in mvl4; outval : inout mvl4);
   end component;
   component xor2_4
      generic (Tmin, Tmax : time);
      port ( inval1, inval2 : in mvl4; outval : inout mvl4);
   end component;
   signal t1,t2,t3,t4,t5 : mvl4;
begin
   -- block of sum
   L11: xor2_4 generic map (Tmin,Tmax) port map (a,b,t1);
   L12: xor2_4 generic map (Tmin,Tmax) port map (t1,cin,sum);
   -- block of carry
   L21: and2_4 generic map (Tmin,Tmax) port map (a,b,t2);
   L22: and2_4 generic map (Tmin,Tmax) port map (a,cin,t3);
   L23: and2_4 generic map (Tmin,Tmax) port map (b,cin,t4);
   L24: or2_4  generic map (Tmin,Tmax) port map (t2,t3,t5);
   L25: or2_4  generic map (Tmin,Tmax) port map (t5,t4,cout);
end struc1;

-- FA with 2 gate stages (using nand3)
-- this is more similiar with real ICs
---------------------------------------
architecture struc2 of full_adder is
   component nand2_4
      generic (Tmin, Tmax : time);
      port ( inval1, inval2 : in mvl4; outval : inout mvl4);
   end component;
   component nand3_4
      generic (Tmin, Tmax : time);
      port ( inval1, inval2, inval3 : in mvl4; outval : inout mvl4);
   end component;
   component xor2_4
      generic (Tmin, Tmax : time);
      port ( inval1, inval2 : in mvl4; outval : inout mvl4);
   end component;
   signal t1,t2,t3,t4 : mvl4;
begin
   -- block of sum
   L11: xor2_4 generic map (Tmin,Tmax) port map (a,b,t1);
   L12: xor2_4 generic map (Tmin,Tmax) port map (t1,cin,sum);
   -- block of carry
   L21: nand2_4 generic map (Tmin,Tmax) port map (a,b,t2);
   L22: nand2_4 generic map (Tmin,Tmax) port map (a,cin,t3);
   L23: nand2_4 generic map (Tmin,Tmax) port map (b,cin,t4);
   L24: nand3_4 generic map (Tmin,Tmax) port map (t2,t3,t4,cout);
end struc2;


------------------------------------------------------------
--  adder_unit
------------------------------------------------------------
use work.system4.all;
entity adder_unit is
   generic (n : positive;          -- size
	    neg_b : boolean;       -- negate data input b
	    Tmin, Tmax : time);    -- gate delay time
   port ( a, b : mvl4_vector(n-1 downto 0);
	  cin : in mvl4;
	  sum : inout mvl4_vector(n-1 downto 0);
	  cout : inout mvl4);
end adder_unit;

architecture struc of adder_unit is
   component invert4
      generic (Tmin, Tmax : time);
      port ( inval : in mvl4; outval : inout mvl4);
   end component;
   component full_adder
      generic (Tmin, Tmax : time);    -- gate delay time
      port (a, b, cin : mvl4;
	    sum, cout : inout mvl4);
   end component;
--   for L14: full_adder use entity work.full_adder(struc1);
--   for L21: full_adder use entity work.full_adder(struc1);
   signal new_b, c : mvl4_vector(n-1 downto 0);
begin
   -- generate substracter
   L10:    for i in 0 to n-1 generate
   L11:       if neg_b=TRUE generate
   Inv:          invert4 generic map (Tmin,Tmax) port map (b(i),new_b(i));
              end generate;
   L18:       if neg_b=FALSE generate
                 new_b(i) <= b(i);
              end generate;
   L13:       if i=0 generate
   FA0:	         full_adder generic map (Tmin,Tmax)
			    port map (a(0),new_b(0),cin,sum(0),c(0));
	      end generate;
   L15:       if i/=0 generate
   FA:	         full_adder generic map (Tmin, Tmax)
			    port map (a(i),new_b(i),c(i-1),sum(i),c(i));
	      end generate;
	   end generate;
   cout <= c(n-1);
end struc;


------------------------------------------------------------
--  unit for producing results
------------------------------------------------------------
use work.system4.all;
entity result_unit is
   generic (n : positive;          -- size
	    Tmin, Tmax : time);    -- gate delay time
   port ( result : mvl4_vector(n-1 downto 0);
	  G,E,L: out bit);
end result_unit;

architecture beh of result_unit is
begin
   process (result)
      variable m : integer;
      variable d_min, d_max : time;
      variable r : mvl4;
   begin
      -- count delay time
      -- delay time = Tmax * log2(n)
      m := 1;
      while 2**m<n loop
	 m := m+1;
      end loop;
      d_min := m * Tmin;
      d_max := m * Tmax;

      if result(n-1)='1' then
	 L <= '1' after d_max;
	 G <= '0' after d_max;
	 E <= '0' after d_max;
      else
	 L <= '0' after d_max;
	 r := '0';
         for i in 0 to n-2 loop
            r := r or result(i);
         end loop;
	 if r='0' then
	    E <= '1' after d_max;
	    G <= '0' after d_max;
	 else
	    G <= '1' after d_max;
	    E <= '0' after d_max;
	 end if;
      end if;
   end process;
end beh;

architecture struc of result_unit is
   component invert4
      generic (Tmin, Tmax : time);
      port ( inval : in mvl4; outval : inout mvl4);
   end component;
   component and2_4
      generic (Tmin, Tmax : time);
      port ( inval1, inval2 : in mvl4; outval : inout mvl4);
   end component;
   component nor2_4
      generic (Tmin, Tmax : time);
      port ( inval1, inval2 : in mvl4; outval : inout mvl4);
   end component;
   component or2_4
      generic (Tmin, Tmax : time);
      port ( inval1, inval2 : in mvl4; outval : inout mvl4);
   end component;
   
   function MVL4toBIT (v: mvl4) return bit;
   function MVL4toBIT (v: mvl4) return bit is
      variable result : bit;
   begin
      case v is
         when '0' => result := '0';
         when '1' => result := '1';
         when 'X' => result := '0';
         when 'Z' => result := '1';
      end case;
      return result;
   end MVL4toBIT;

   signal r : mvl4_vector (0 to 2*n-2);
   signal t1,t3,t4 : mvl4;
begin
   L1: for i in 0 to n-1 generate
          r(i) <= result(i);
       end generate;

   L2: for i in 0 to n-2 generate
   Or2: or2_4 generic map (Tmin, Tmax) port map (r(2*i),r(2*i+1),r(n+i));
       end generate;

   L <= MVL4toBIT(result(n-1));

   E1: nor2_4 generic  map (Tmin, Tmax) port map (result(n-1),r(2*n-2),t1);
   E <= MVL4toBIT (t1);

   G1: invert4 generic map (Tmin, Tmax) port map (result(n-1),t3);
   G2: and2_4 generic  map (Tmin, Tmax) port map (t3,r(2*n-2),t4);
   G <= MVL4toBIT (t4);
 end struc;


------------------------------------------------------------
--  simple comparator
------------------------------------------------------------
entity comparator is
    generic (n : positive;          -- size
	     Tmin, Tmax : time);    -- gate delay time
    port ( A,B : bit_vector(n-1 downto 0);
	   G,E,L: out bit);
end comparator;


------------------------------------------------------------
--  pipelined comparator
------------------------------------------------------------
use work.system4.all;
entity pipelined_comparator is
   generic (n : positive;          -- size of comparator
            unit : positive;       -- size of each adder unit
	    Tmin, Tmax : time);    -- gate delay time
   port (A,B : bit_vector(n-1 downto 0);
	 clk: bit;
	 G,E,L: out bit);
end pipelined_comparator;


architecture struc of pipelined_comparator is
   component latch
      generic (n : positive;          -- size
	       Tmin, Tmax : time);    -- gate delay time
      port ( cin : in mvl4_vector(n-1 downto 0);
	     clk : in mvl4;
	     cout : inout mvl4_vector(n-1 downto 0) );
   end component;

   component adder_unit
      generic (n : positive;          -- size
	       neg_b: boolean;        -- negate input data b
	       Tmin, Tmax : time);    -- gate delay time
      port ( a, b : mvl4_vector(n-1 downto 0);
	     cin : mvl4;
	     sum : inout mvl4_vector(n-1 downto 0);
	     cout : inout mvl4);
   end component;
   -- for Ladd: adder_unit use entity work.adder_unit(struc);

   component result_unit
      generic (n : positive;          -- size
	       Tmin, Tmax : time);    -- gate delay time
      port ( result : mvl4_vector(n-1 downto 0);
	     G,E,L: out bit);
   end component;

   -- Bit to Vector
   function BtoV (b: mvl4) return mvl4_vector;
   function BtoV (b: mvl4) return mvl4_vector is
      variable result : mvl4_vector(0 to 0);
   begin 
      result(0):=b;
      return result; 
   end;


   constant m : integer := n/unit;
   type input_stream is array (0 to m-1, m-1 downto 0) of mvl4_vector(unit-1 downto 0);
   signal Ain, Bin : input_stream;
   type sum_stream is array (0 to m, m-1 downto 0) of mvl4_vector(unit-1 downto 0);
   signal sum : sum_stream;
   type carry_stream is array (0 to m-1) of mvl4;
   signal Cin, Cout: carry_stream;
   signal Sout : mvl4_vector (n-1 downto 0);
   signal sys_clk : mvl4;

begin
   sys_clk <= BITtoMVL4 (clk) after 1 ns;

   -- generate latches for each column
   ----------------------------------------
   L1:  for col in 0 to m-1 generate

           -- generate upper latches
   L11:	   for row in 0 to col generate
   L12:       if row=0 generate
   L0A:         latch generic map (unit,Tmin,Tmax) port map 
                     (BVtoMVL4V(A(unit*col+(unit-1) downto unit*col)),sys_clk,Ain(row,col));
   L0B:         latch generic map (unit,Tmin,Tmax) port map 
                     (BVtoMVL4V(B(unit*col+(unit-1) downto unit*col)),sys_clk,Bin(row,col));
	      end generate;
   L13:       if row/=0 generate
   LA:          latch generic map (unit,Tmin,Tmax)
		       port map (Ain(row-1,col),sys_clk,Ain(row,col));
   LB:          latch generic map (unit,Tmin,Tmax)
		       port map (Bin(row-1,col),sys_clk,Bin(row,col));
	      end generate;
           end generate;

           -- initialize all not used signal 'Z' for debugging
   L14:    for row in col+1 to m-1 generate
   L15:       for k in unit-1 downto 0 generate
                 Ain(row,col)(k) <= 'Z';
                 Bin(row,col)(k) <= 'Z';
              end generate;
           end generate;

           -- generate lower latches
   L16:	   for row in col+1 to m generate
   LS:       latch generic map (unit,Tmin,Tmax)
		    port map (sum(row-1,col),sys_clk,sum(row,col));
	   end generate;

           -- initialize all not used signal 'Z' for debugging
   L17:    for row in 0 to col-1 generate
   L18:       for k in unit-1 downto 0 generate
                 sum(row,col)(k) <= 'Z';
              end generate;
           end generate;
	end generate;  -- latches

   -- generate adder_units
   L2:  for i in 0 to m-1 generate
   Add:    adder_unit generic map (unit,TRUE,Tmin,Tmax)
		      port map (Ain(i,i),Bin(i,i),Cin(i),sum(i,i),Cout(i));
   L22:    if i<m-1 generate   -- carry latches
   LC:       latch generic map (1,Tmin,Tmax) 
                    port map (btov(Cout(i)),sys_clk,btov(Cin(i+1)));
	   end generate;
	end generate;
	Cin(0) <= '1';

   -- assign result to bit_vector Rout
   process (sum)
   begin
      for i in m-1 downto 0 loop
	 Sout(unit*i+(unit-1) downto unit*i) <= sum(m,i);
      end loop;
   end process;

   Rult: result_unit generic map (n,Tmin,Tmax) port map (Sout,G,E,L);
end struc;


------------------------------------------------------------
--  test bench of comparator
------------------------------------------------------------
entity test_comparator is end;

architecture struc of test_comparator is
   constant n : positive := 24;      -- size of comparator
   constant unit : positive :=2;     -- size of eacher adder unit
   constant dat_num : integer := 20;
   constant Tmin : time := 3 ns;
   constant Tmax : time := 6 ns;

   component pipelined_comparator
      generic (n : positive;          -- size of comparator
               unit : positive;       -- size of each adder unit
	       Tmin, Tmax : time);    -- gate delay time
      port (A,B : bit_vector(n-1 downto 0);
	    clk: bit;
	    G,E,L: out bit);
   end component;

   signal clk, G,E,L : bit;
   signal A, B : bit_vector (n-1 downto 0);
   type data_list is array (0 to dat_num-1) of bit_vector (n-1 downto 0);
   constant dat_A : data_list :=
    ( B"000000000000000000000000",
      B"011111111111111111111111",
      B"000000000000000000000011",
      B"000000000000000000000100",
      B"100000000000000000000000",
      B"011111111111111111111111",
      B"000000000000000011000000",
      B"010000000000000000000000",
      B"101010101010101010101010",
      B"010101010101010101010101",
      others => B"000000000000000000000000");
   constant dat_B : data_list :=
    ( B"000000000000000000000000",
      B"100000000000000000000000",
      B"000000000000000000000110",
      B"000000000000000000000100",
      B"101010101010101010101010",
      B"000000000001000000110000",
      B"111111111111111111111111",
      B"100000000000000000000000",
      B"101000001000000000000000",
      B"100100000000011010111000",
      others => B"000000000000000000000000");
begin
   clk <= not clk after 22 ns;    -- pipeline cycle =  ns;

   PC: pipelined_comparator generic map(n,unit,Tmin,Tmax) 
                            port map(A,B,clk,G,E,L);

   process (clk)
      variable i : integer := 0;
   begin
      if clk'event and clk='1' then
	 A <= dat_A(i);
	 B <= dat_B(i);
	 if i<dat_num-1 then
            i := i+1;
         end if;
      end if;
   end process;
end struc;


------------------------------------------------------------
--  test bench of adder_unit
------------------------------------------------------------
use work.system4.all;
entity test_adder is end;

architecture struc of test_adder is
   constant n : positive := 4;
   component adder_unit
      generic (n : positive;          -- size
	       neg_b: boolean;        -- negate input data b
	       Tmin, Tmax : time);    -- gate delay time
      port ( a, b : mvl4_vector(n-1 downto 0);
	     cin : mvl4;
	     sum : inout mvl4_vector(n-1 downto 0);
	     cout : inout mvl4);
   end component;
--   for L1: adder_unit use entity work.adder_unit(beh);
--   for L2: adder_unit use entity work.adder_unit(struc);
   signal a,b,sum,sub : mvl4_vector(n-1 downto 0);
   signal cin1,cout1, cin2,cout2 : mvl4;
begin
   L1: adder_unit generic map (4,FALSE,3 ns,6 ns)
		  port map (a,b,cin1,sum,cout1);
   L2: adder_unit generic map (4,TRUE,3 ns,6 ns)
		  port map (a,b,cin2,sub,cout2);

   process
   begin
      cin1 <= '0';
      cin2 <= '1';

      a <= "0001";
      b <= "0010";
      wait for 50 ns;

      a <= "0110";
      b <= "0001";
      wait for 50 ns;

      a <= "1111";
      b <= "1111";
      wait for 50 ns;

      wait;
   end process;
end struc;
