
entity count_4 is
    type level_2_vector is array (integer range <>) of bit;
end count_4;
     --
     --  implement this counter from boolean equations for a
     --  asynchronous scale of 4 counter. The corresponding
     --  status table is like following:
     --
     -- sig(input)|    0     |    1 
     -- ----------|----------|------------
     --  S0 (000) | 0 0 0  0 | 0 0 1  0
     --  S1 (001) | 0 1 1  0 | 0 0 1  0
     --  S2 (011) | 0 1 1  0 | 0 1 0  0
     --  S3 (010) | 1 1 0  0 | 0 1 0  0
     --  S4 (110) | 1 1 0  0 | 1 1 1  0
     --  S5 (111) | 1 0 1  0 | 1 1 1  0
     --  S6 (101) | 1 0 1  0 | 1 0 0  1
     --  S7 (100) | 0 0 0  0 | 1 0 0  1
     --
     -- The Karnaugh map of s0, s1, s2 and sig4
     -- 
     -- s0 sig |
     -- ----+  |     (s0)    |     (s1)    |
     -- s1s2+-+| 00 01 11 10 | 00 01 11 10 |
     -- ------++-------------+-------------+
     --  0  0  |  0  0  1  0 |  0  0  0  0 |
     --  0  1  |  0  0  1  1 |  1  0  0  0 |
     --  1  1  |  0  0  1  1 |  1  1  1  0 |
     --  1  0  |  1  0  1  1 |  1  1  1  1 |
     --         
     -- s0 sig |
     -- ----+  |     (s2)    |    (sig4)   |   
     -- s1s2+-+| 00 01 11 10 | 00 01 11 10 |   
     -- ------++-------------+-------------+
     --  0  0  |  0  1  0  0 |  0  0  1  0 |   
     --  0  1  |  1  1  0  1 |  0  0  1  0 |   
     --  1  1  |  1  0  1  1 |  0  0  0  0 |   
     --  1  0  |  0  0  1  0 |  0  0  0  0 |   
     --
     --  now we get the boolean equations by  Karnaug map
     --
     --  s0 <= (s0 and s2) or (s0 and (not s2) and sig)
     --       (s1 and (not s2) and (not sig));
     --       
     --  s1 <= (s1 and (not s2) and (not sig)) or
     --       (s1 and sig) or ((not s0) and s2 and (not sig));
     -- 
     --  s2 <= ((not s0) and s2 and (not sig)) or
     --       (s0 and s2 and (not sig)) or (s0 and s1 and sig) or 
     --       ((not s0) and (not s1) and sig);
     -- 
     --  sig4 <= s0 and (not s1) and sig;

architecture beh of count_4 is
  signal sig, s0, s1, s2, sig4: bit;
  signal sv : level_2_vector (0 to 4);
  signal t : time := 0 us;
begin      
     sv <= sig & s0 & s1 & s2  & sig4;
    
     s0 <= (s0 and s2) or (s0 and (not s2) and sig) or
              (s1 and (not s2) and (not sig));
            
     s1 <= (s1 and (not s2) and (not sig)) or
           (s1 and sig) or ((not s0) and s2 and (not sig));

     s2 <= ((not s0) and s2 and (not sig)) or
           (s0 and s2 and (not sig)) or (s0 and s1 and sig) or 
           ((not s0) and (not s1) and sig);

     sig4 <= s0 and (not s1) and sig;

     sig <= not sig after 10 ns;
     t <= t + 10 ns after 10 ns;
     assert t < 300 ns
           report "overtime"
           severity error;
end;







entity count_3 is
    type level_2_vector is array (integer range <>) of bit;
end count_3;
     --
     --  implement this counter from boolean equations for a
     --  asynchronous scale of 4 counter. The corresponding
     --  status table is like following:
     --
     -- sig(input)|    0     |    1 
     -- ----------|----------|------------
     --  S0 (000) | 0 0 0  0 | 0 0 1  0
     --  S1 (001) | 0 1 1  0 | 0 0 1  0
     --  S2 (011) | 0 1 1  0 | 0 1 0  0
     --  S3 (010) | 1 1 0  0 | 0 1 0  0
     --  S4 (110) | 1 1 0  0 | 1 1 1  d
     --  S5 (111) | 0 0 0  0 | 1 1 1  1
     --
     -- The Karnaugh map of s0, s1, s2 and sig4
     -- 
     -- s0 sig |
     -- ----+  |     (s0)    |     (s1)    |
     -- s1s2+-+| 00 01 11 10 | 00 01 11 10 |
     -- ------++-------------+-------------+
     --  0  0  |  0  0  d  d |  0  0  d  d |
     --  0  1  |  0  0  d  d |  1  0  d  d |
     --  1  1  |  0  0  1  0 |  1  1  1  d |
     --  1  0  |  1  0  1  1 |  1  1  1  1 |
     --         
     -- s0 sig |
     -- ----+  |     (s2)    |    (sig4)   |   
     -- s1s2+-+| 00 01 11 10 | 00 01 11 10 |   
     -- ------++-------------+-------------+
     --  0  0  |  0  1  d  d |  0  0  d  d |   
     --  0  1  |  1  1  d  d |  0  0  d  d |   
     --  1  1  |  1  0  1  0 |  0  0  1  0 |   
     --  1  0  |  0  0  1  0 |  0  0  0  0 |   
     --
     --  now we get the boolean equations by  Karnaug map
     --
     --  s0 <= (s0 and (sig or (not s2)) or 
     --       (s1 and (not s2) and (not sig));
     --       
     --  s1 <= (s1 and (not s2)) or (sig and s1) or (s2 and (not(sig or s0)))
     -- 
     --  s2 <= (sig and (not s1)) or (sig and s0) or 
     --         ((not s0) and (not sig) and s2);
     -- 
     --  sig4 <= s0 and s1 and sig;

architecture beh of count_3 is
  signal sig, s0, s1, s2, sig4: bit;
  signal sv : level_2_vector (0 to 4);
  signal t : time := 0 us;
begin      
     sv <= sig & s0 & s1 & s2  & sig4;
    
       s0 <= (s0 and (sig or (not s2))) or 
            (s1 and (not s2) and (not sig));
            
       s1 <= (s1 and (not s2)) or (sig and s1) or (s2 and (not(sig or s0)));
      
       s2 <= (sig and (not s1)) or (sig and s0) or 
              ((not s0) and (not sig) and s2);
    
      sig4 <=  s0 and s1  and sig;
     

     sig <= not sig after 10 ns;
     t <= t + 10 ns after 10 ns;
     assert t < 300 ns
           report "overtime"
           severity error;
end;






