------------------------------------------------------------------------
--  Object:  CSE 561 Term Project
--  Title :  Powerful elevator for a building with any stories
--  Author:  Chung-Cheng Kuo
--  Date  :  11/03/1994
------------------------------------------------------------------------
--  Introduction:
--    1. Suitable for any stories, even basements
--    2. With OPEN and CLOSE button
--    3. Can display FLOOR LIGHT
--    4. Use text to reprent signals while simulating
--    5. Intellegent, stop only for needed
--
--  Control Rules:
--    1. If no one use it, stop to waiting
--    2. Stay upward or downward if there are any requests
--    3. Backward as soon as no moving needed
--    4. If the OPEN button stay pressed, the elevator will wait for
--       the door closed.
--    5. If the up/down, or floor buttons stay pressed, 
--       it can also work but stop more.
------------------------------------------------------------------------

entity elevator is 
end elevator;

architecture beh of elevator is

  constant LOWEST_FLOOR  : integer := 0;        -- basement B1 --> 0
  constant HIGHEST_FLOOR : integer := 5;

  -- Define the resolution function
  function bit_or(invals : bit_vector) return bit is
    variable result : bit := '0';
    begin
      for i in invals'range loop
       result := result or invals(i);
      end loop;
      return result;
  end bit_or;

  -- Define the resolved type, res_bit;
  subtype res_bit is bit_or bit;
  type res_bit_vector is array(integer range <>) of res_bit;

  -- Define specification types of th elevator
  type FLOOR_RANGE is range LOWEST_FLOOR to HIGHEST_FLOOR;
  type MOTOR_CONTROLS is (DOWN, STOP, UP);
  type DOOR_CONTROLS  is (CLOSE_DOOR, OPEN_DOOR);
  type DOOR_POSITIONS is (CLOSED, OPENED);

  signal position      : integer := 10;   -- first floor
  signal door_position : DOOR_POSITIONS := CLOSED;  
  signal door_control  : DOOR_CONTROLS  := CLOSE_DOOR; 
  -- Make the buttons and lights resolved signals
  signal floor_buttons, up_buttons, down_buttons,
         floor_lights,  up_lights,  down_lights : 
                                  res_bit_vector(LOWEST_FLOOR to HIGHEST_FLOOR);
  signal close_button  : res_bit := '0';   -- button to force door to close
  signal open_button   : res_bit := '0';   -- button to force door to open
  signal position_light: bit_vector(LOWEST_FLOOR to HIGHEST_FLOOR);
  signal sweep_light   : MOTOR_CONTROLS := STOP;
  signal motor_control : MOTOR_CONTROLS := STOP;  
  signal upward_served, downward_served : bit := '0';
                                  -- controller use them to clear up/down lights
  signal tm : time := 0 sec;


  -- This function count the floor number with position
  ----------------------------------------------------------
  function curr_floor (position: integer) return integer is
  begin
     if (position mod 10=0) then
        return position/10;
     else
        return -100;
     end if;
  end curr_floor;


  -- This function count the new direction of the elevator while stoping at or
  --   going over a floor
  ------------------------------------------------------------------------------
  function update_sweep_light (sweep_light: MOTOR_CONTROLS;  position: integer;
           up_lights, down_lights, floor_lights : 
                                 res_bit_vector(LOWEST_FLOOR to HIGHEST_FLOOR))
           return MOTOR_CONTROLS is
     variable new_sweep: MOTOR_CONTROLS;
     variable floor,i : integer;
     variable need_up,need_down : res_bit;
  begin
     floor := curr_floor(position);

     -- Check if need to serve the higher floors
        need_up := '0';
        for i in floor+1 to HIGHEST_FLOOR loop
           need_up := need_up or up_lights(i);
           need_up := need_up or down_lights(i);
           need_up := need_up or floor_lights(i);
        end loop;

     -- Check if need to serve the lower floors
        need_down := '0';
        for i in LOWEST_FLOOR to floor-1 loop
           need_down := need_down or up_lights(i);
           need_down := need_down or down_lights(i);
           need_down := need_down or floor_lights(i);
        end loop;

     -- Judge the new direction
        if (need_down='1') then    -- lower need service
           if (sweep_light=UP and need_up='1') then
              new_sweep := UP;   -- upward
           else
              new_sweep := DOWN;    -- downward
           end if;
        else
           if (need_up='1') then
              new_sweep := UP;    -- upward
           else
              new_sweep := STOP;    -- stop to wait for service
           end if;
        end if;
     return new_sweep;
  end update_sweep_light;


begin    -- of the architecher of elevator

    -- This process models the way the elevator position is affected by the motor
    process
       -- I intend to initialize position to 10 (first floor)
       -- and motor_control to STOP
       begin
            wait on motor_control, position;  -- ?? position
            if (motor_control /= STOP) then
               position <= position + MOTOR_CONTROLS'pos(motor_control)-1 
                                                                  after 0.2 sec;
            else
               position <= position;              -- to cancel scheduled changes
            end if;

         assert (position <= HIGHEST_FLOOR*10)
            report "Blast off"
            severity error;

         assert (position >= LOWEST_FLOOR*10)
            report "Position too low!"
            severity error;
    end process;


    -- This process models the way the door position
    -- depends on the door_control and door open_button
    ---------------------------------------------------------------------    
    process (door_control, open_button)
       begin
          if (open_button='1' and (position mod 10)=0 and motor_control=STOP) then
             if door_position=CLOSED then
                door_position <= OPENED after 1 sec;
             end if;
          else
             door_position <= DOOR_POSITIONS'val(DOOR_CONTROLS'pos(door_control))
                                                                     after 1 sec;
          end if;
       end process;


    -- This process trace the up_buttons
    ---------------------------------------------------------------------
    process (up_buttons, upward_served)
       variable i: integer;
    begin
       for i in LOWEST_FLOOR to HIGHEST_FLOOR-1 loop
          if up_buttons(i)='1' then up_lights(i) <= '1'; end if;
       end loop;
       if upward_served='1' then
          up_lights(curr_floor(position)) <= '0';
       end if;
    end process;


    -- This process trace the down_buttons
    ---------------------------------------------------------------------
    process (down_buttons, downward_served)
       variable i: integer;
    begin
       for i in LOWEST_FLOOR+1 to HIGHEST_FLOOR loop
          if down_buttons(i)='1' then down_lights(i) <= '1'; end if;
       end loop;
       if downward_served='1' then
          down_lights(curr_floor(position)) <= '0';
       end if;
    end process;


    -- This process trace the floor_buttons
    ---------------------------------------------------------------------
    process (floor_buttons, door_position)
       variable i: integer;
    begin
       for i in LOWEST_FLOOR to HIGHEST_FLOOR loop
          if floor_buttons(i)='1' then floor_lights(i) <= '1'; end if;
       end loop;
       if door_position=OPENED then
          floor_lights(curr_floor(position)) <= '0';
       end if;
    end process;


    ------------------------------------------------------------------------
    --  This process models my elevator control system
    ------------------------------------------------------------------------
    process
       variable i,floor: integer;
       variable need_up, need_down : res_bit;
       variable old_sweep, new_sweep: MOTOR_CONTROLS;
       variable rr1, rr2:bit;
    begin
       if ((position mod 10) = 0) then       -- if reach a floor
          -- Now the elevator is at a floor but stopped or not
          floor:= curr_floor(position);    

          -- Set the position lights
          for i in LOWEST_FLOOR to HIGHEST_FLOOR loop
              position_light(i) <= '0';
          end loop;
          position_light(floor) <= '1';

          -- Count the need for upward service
          need_up := '0';
          for i in floor+1 to HIGHEST_FLOOR loop
             need_up := need_up or up_lights(i);
             need_up := need_up or down_lights(i);
             need_up := need_up or floor_lights(i);
          end loop;
          need_up := need_up or up_lights(floor);

          -- Count the need for downward service
          need_down := '0';
          for i in LOWEST_FLOOR to floor-1 loop
             need_down := need_down or up_lights(i);
             need_down := need_down or down_lights(i);
             need_down := need_down or floor_lights(i);
          end loop;
          need_down := need_down or down_lights(floor);

          if ( floor_lights(floor)='1' or            -- need service at this floor
              (up_lights(floor)='1' and (sweep_light/=DOWN or need_down='0')) or
              (down_lights(floor)='1' and (sweep_light/=UP or need_up='0') )) then

             -- Stop to serve a floor
             motor_control <= STOP;
             door_control <= OPEN_DOOR;
             if (door_position /= OPENED) then
                wait until door_position = OPENED;
             end if;
             wait until close_button='1' for 10 sec;
             door_control <= CLOSE_DOOR;
             wait until door_position = CLOSED;
             new_sweep := update_sweep_light(sweep_light, position, up_lights,
                                             down_lights, floor_lights);
             case new_sweep is
                when UP   => upward_served   <= '1', '0' after 1 sec;
                when DOWN => downward_served <= '1', '0' after 1 sec;
                when STOP => 
             end case;
             sweep_light <= new_sweep;
             motor_control <= new_sweep;
          else    -- needn't service at this floor (go over)
             old_sweep := sweep_light;
             new_sweep := update_sweep_light(sweep_light, position, up_lights,
                                             down_lights, floor_lights);
             -- stop to reverse the direction if needed
             if (old_sweep/=new_sweep and motor_control/=STOP) then  
                motor_control <= STOP;   
                wait for 1 sec;
             end if;
             sweep_light <= new_sweep;
             motor_control <= new_sweep;
          end if;
       end if;
       wait on position for 1 sec;
    end process;


    -- Security check while the elevator is between two floors
    ----------------------------------------------------------------------
    assert (door_position = CLOSED or motor_control = STOP)
       report "Door opened while the elevator is moving!"
       severity error;


    -- This process models the behavior of someone wanting
    -- to go from floor 1 to floor 5
    ----------------------------------------------------------------------
    process -- person 1
      begin
        wait for 1 sec;
        up_buttons(1) <= '1', '0' after 0.5 sec;
        if not(curr_floor(position)=1 and door_position=OPENED) then
           wait until (curr_floor(position)=1 and door_position=OPENED);
        end if;
        floor_buttons(5) <= '1', '0' after 0.5 sec;
         wait until curr_floor(position)=5 and door_position=OPENED;
        wait;
    end process;


    -- This process models the behavior of someone wanting
    -- to go from floor 2 to floor 4
    ----------------------------------------------------------------------
    process -- person 2
      begin
        wait for 1 sec;
        up_buttons(2) <= '1', '0' after 0.5 sec;
        if not(curr_floor(position)=2 and door_position=OPENED) then
           wait until (curr_floor(position)=2 and door_position=OPENED);
        end if;
        open_button <= '1', '0' after 20 sec;
        floor_buttons(4) <= '1', '0' after 0.5 sec;
        wait until curr_floor(position)=4 and door_position=OPENED;
        wait;
    end process;


    -- This process models the behavior of someone wanting
    -- to go from floor 3 to floor B1
    ----------------------------------------------------------------------
    process
      begin
        wait for 5 sec;
        down_buttons(3) <= '1', '0' after 0.5 sec;
        if not(curr_floor(position)=3 and door_position = OPENED) then
           wait until curr_floor(position)=3 and door_position=OPENED;
        end if;
        floor_buttons(0) <= '1', '0' after 0.5 sec;
        close_button <= '1', '0' after 0.5 sec;
        wait until curr_floor(position)=0 and door_position=OPENED;
        wait;
    end process;


    -- This process models the behavior of someone wanting
    -- to go from floor 5 to floor 1
    ----------------------------------------------------------------------
    process
      begin
        wait for 10 sec;
        down_buttons(5) <= '1', '0' after 0.5 sec;
        if not(curr_floor(position)=5 and door_position=OPENED) then
           wait until (curr_floor(position)=5 and door_position=OPENED);
        end if;
        floor_buttons(1) <= '1', '0' after 0.5 sec;
        close_button <= '1', '0' after 0.5 sec;
        wait until curr_floor(position)=1 and door_position=OPENED;
        wait;
    end process;


    -- This process models the behavior of IRREGULAR conditions
    -- from floor 2 to 1 and 3, but press up_buttons(2) and 
    --   floor_buttons(1) forever
    ----------------------------------------------------------------------
    process
      begin
        wait for 15 sec;
        up_buttons(2) <= '1';                        -- PUSH UP_BUTTON FOREVER
        if not(curr_floor(position)=2 and door_position=OPENED) then
           wait until (curr_floor(position)=2 and door_position=OPENED);
        end if;
        floor_buttons(3) <= '1', '0' after 0.5 sec;
        floor_buttons(1) <= '1';                     -- PUSH FLOOR_BUTTON FOREVER
        close_button <= '1', '0' after 0.5 sec;
        wait until curr_floor(position)=1 and door_position=OPENED;
        wait;
    end process;

end beh;

