
entity elevator is end elevator;

architecture beh of elevator is

constant STORIES       : integer := 4;
constant LOWEST_FLOOR  : integer := 1;        -- basement B1 --> 0
constant HIGHEST_FLOOR : integer := 4;

  -- Define the resolution function
  function bit_or(invals : bit_vector) return bit is
    variable result : bit := '0';
    begin
      for i in invals'range loop
       result := result or invals(i);
      end loop;
      return result;
  end bit_or;

  -- Define the resolved type, res_bit;
  subtype res_bit is bit_or bit;
  type res_bit_vector is array(integer range <>) of res_bit;

  -- Define specification types of th elevator
  type FLOOR_RANGE is range LOWEST_FLOOR to HIGHEST_FLOOR;
  type MOTOR_CONTROLS is range -1 to 1;
  type DOOR_CONTROLS  is (CLOSE_DOOR, OPEN_DOOR);
  type DOOR_POSITIONS is (CLOSED, OPENED);

  signal position      : integer := 10;  -- first floor
  signal door_position : DOOR_POSITIONS := CLOSED;   -- closed
  signal door_control  : DOOR_CONTROLS  := CLOSE_DOOR;   -- close
  -- Make the buttons resolved signals
  signal floor_buttons, up_buttons, down_buttons: bit_vector(LOWEST_FLOOR to HIGHEST_FLOOR);
  signal floor_lights,  up_lights,  down_lights : res_bit_vector(LOWEST_FLOOR to HIGHEST_FLOOR);
  signal close_button  : bit := '0';   -- force door close
  signal open_button   : bit := '0';   -- force door open
  signal position_light: bit_vector(LOWEST_FLOOR to HIGHEST_FLOOR);
  signal sweep_light   : MOTOR_CONTROLS := 0;
  signal motor_control : MOTOR_CONTROLS := 0;   -- stopped
  signal tm : time := 0 sec;

  function curr_floor (position: integer) return integer is
  begin
     if (position mod 10=0) then 
        return position/10;
     else 
        return -100;
     end if;
  end curr_floor;

  function need_service (up_lights, down_lights, floor_lights :
     res_bit_vector(LOWEST_FLOOR to HIGHEST_FLOOR)) return bit is
     variable i: integer;
     variable result: res_bit := '0';
  begin
     for i in LOWEST_FLOOR to HIGHEST_FLOOR loop
         result := result or up_lights(i);
         result := result or down_lights(i);
         result := result or floor_lights(i);
     end loop;
     return result;
  end need_service;

  function update_sweep_light (sweep_light: MOTOR_CONTROLS; position: integer;
           up_lights, down_lights, floor_lights : res_bit_vector(LOWEST_FLOOR to HIGHEST_FLOOR))
           return MOTOR_CONTROLS is
     variable new_sweep: MOTOR_CONTROLS;
  begin
     case position/10 is
        when 1 => if (need_service(up_lights,down_lights,floor_lights)='1') then
                     new_sweep := 1;  -- up
                  else
                     new_sweep := 0;  -- stop
                  end if;
        when 4 => if (need_service(up_lights,down_lights,floor_lights)='1') then
                     new_sweep := -1;  -- down
                  else
                     new_sweep := 0;   -- stop
                  end if;

        when 2 => if (up_lights(3) or up_lights(4) or   -- 1 & 2 need service
                      down_lights(3) or down_lights(4) or
                      floor_lights(3) or floor_lights(4))='1' then
                     if (sweep_light=1 or (up_lights(1) or
                                     down_lights(1) or floor_lights(1))='0') then
                        new_sweep := 1;  -- up
                     else
                        new_sweep := -1;  -- down
                     end if;
                  else
                     if (up_lights(1) or down_lights(1) or floor_lights(1))='1' then
                        new_sweep := -1;  -- down
                     else
                        new_sweep := 0;  -- waiting for serving
                     end if;
                  end if;

        when 3 => if (up_lights(1) or up_lights(2) or   -- 1 & 2 need service
                      down_lights(1) or down_lights(2) or
                      floor_lights(1) or floor_lights(2))='1' then
                     if (sweep_light=-1 or (up_lights(4) or
                                     down_lights(4) or floor_lights(4))='0') then
                        new_sweep := -1;  -- down
                     else
                        new_sweep := 1;  -- up
                     end if;
                  else
                     if (up_lights(4) or down_lights(4) or floor_lights(4))='1' then
                        new_sweep := 1;  -- up
                     else
                        new_sweep := 0;  -- waiting for serving
                     end if;
		  end if;
        when others => new_sweep:=1; -- report "Incorrect position.";
                       -- severity error;

        end case;
     return new_sweep;
  end update_sweep_light;

begin

  -- This process models the way the elevator position
  -- is affected by the motor
      process
       -- I intend to initialize position to 10
       -- and motor_control to 0
         begin
              wait on motor_control, position;  -- ?? position
              if (motor_control /= 0) then
                 position <= position + MOTOR_CONTROLS'pos(motor_control) after 200 ms;
              else
                 position <= position;  -- to cancel scheduled changes
              end if;

           assert (position <= HIGHEST_FLOOR*10)
              report "Blast off"
              severity error;

           assert (position >= LOWEST_FLOOR*10)
              report "Position too low!"
              severity error;
      end process;


   -- This process models the way the door position
   -- depends on the door_control and door open_button
      process (door_control, open_button)
         begin
            if (open_button='1' and (position mod 10)=0 and motor_control=0) then
               if door_position=CLOSED then 
                  door_position <= OPENED after 1 sec;
               end if;
            else
               door_position <= DOOR_POSITIONS'val(DOOR_CONTROLS'pos(door_control)) after 1 sec;
            end if;
         end process;


    -- The following 3 processes trace the up_buttons, down_buttons,
    --     and floor_buttons

    process (up_buttons, door_position)
       variable i: integer;
    begin
       for i in LOWEST_FLOOR to HIGHEST_FLOOR loop
          if up_buttons(i)='1' then up_lights(i) <= '1'; end if;
       end loop;
       if door_position=OPENED and sweep_light=1 then
          up_lights(curr_floor(position)) <= '0';
       end if;
    end process;

    process (down_buttons, door_position)
       variable i: integer;
    begin
       for i in LOWEST_FLOOR to HIGHEST_FLOOR loop
          if down_buttons(i)='1' then down_lights(i) <= '1'; end if;
       end loop;
       if door_position=OPENED and sweep_light=-1 then
          down_lights(curr_floor(position)) <= '0';
       end if;
    end process;

    process (floor_buttons, door_position)
       variable i: integer;
    begin
       for i in LOWEST_FLOOR to HIGHEST_FLOOR loop
          if floor_buttons(i)='1' then floor_lights(i) <= '1'; end if;
       end loop;
       if door_position=OPENED then
          floor_lights(curr_floor(position)) <= '0';
       end if;
    end process;


    -- The following models my elevator control system
    process
       variable i: integer;
       variable old_sweep, new_sweep: MOTOR_CONTROLS;
    begin
       if ((position mod 10) = 0) then
          for i in LOWEST_FLOOR to HIGHEST_FLOOR loop
              position_light(i) <= '0';
          end loop;
          i:= position/10;    -- the elevator is at a floor i (stopped or not)
          position_light(i) <= '1';
          if (floor_lights(i)='1' or (sweep_light=1 and up_lights(i)='1') or (sweep_light=-1 and down_lights(i)='1')) then  -- need service at this floor
             motor_control <= 0;
             door_control <= OPEN_DOOR;
             if (door_position /= OPENED) then
                wait until door_position = OPENED;
             end if;
             wait until close_button='1' for 10 sec;
             door_control <= CLOSE_DOOR;
             wait until door_position = CLOSED;
             new_sweep := update_sweep_light(sweep_light,position,up_lights,down_lights,floor_lights);
             sweep_light <= new_sweep;
             motor_control <= new_sweep;
          else
             old_sweep := sweep_light;
             new_sweep := update_sweep_light(sweep_light,position,up_lights,down_lights,floor_lights);
             if (old_sweep/=new_sweep and motor_control/=0) then  -- reverse
                motor_control <= 0;    -- stop
                wait for 1 sec;
             end if;
             sweep_light <= new_sweep;
             motor_control <= new_sweep;
          end if;
       end if;
       wait on position for 1 sec;
    end process;

    assert (door_position = CLOSED or motor_control = 0)
       report "Door opened while the elevator is moving!"
       severity error;


      -- This process models the behavior of someone wanting
      -- to go from floor 1 to floor 3
        process -- person 1
          begin
            wait for 1 sec;
            up_buttons(1) <= '1', '0' after 0.5 sec;
            if not(curr_floor(position)=1 and door_position=OPENED) then
               wait until (curr_floor(position)=1 and door_position=OPENED);
            end if;
            floor_buttons(3) <= '1', '0' after 0.5 sec;
            wait until curr_floor(position)=3 and (door_position=OPENED);
            wait;
        end process;

      -- this process models the behavior of someone wanting
      -- to go from floor 2 to floor 4
        process -- person 2
          begin
            wait for 1 sec;
            up_buttons(2) <= '1', '0' after 0.5 sec;
            if not(curr_floor(position)=2 and (door_position=OPENED) )
              then wait until (curr_floor(position)=2 and (door_position=OPENED));
            end if;
            floor_buttons(4) <= '1', '0' after 0.5 sec;
            wait until curr_floor(position)=4 and (door_position=OPENED);
            wait;
        end process;

      -- this process models the behavior of someone wanting
      -- to go from floor 3 to floor 2
        process
          begin
            wait for 2 sec;
            down_buttons(3) <= '1', '0' after 0.5 sec;
            if not(( door_position = OPENED) and (position = 30))
              then wait until ((door_position = OPENED) and (position = 30));
            end if;
            floor_buttons(2) <= '1', '0' after 0.5 sec;
            wait until curr_floor(position)=2 and (door_position=OPENED);
            wait;
        end process;

end beh;

