package system4 is
 type mvl4 is ('X','0','1','Z');
 type mvl4_vector is array(natural range <>) of mvl4;
 type mvl4_tab1d is array (mvl4) of mvl4;
 type mvl4_table is array (mvl4,mvl4) of mvl4;
 type mvl4_tab3d is array (mvl4,mvl4,mvl4) of mvl4;

 -- table for not
 constant table_not4 :mvl4_tab1d := ('X','1','0','0'); --NOTICE!!!

 constant table_and4 :mvl4_table :=
 -- table for and
 --    X    0    1    Z
 -------------------------
   (( 'X', '0', 'X', 'X' ),  -- X
    ( '0', '0', '0', '0' ),  -- 0
    ( 'X', '0', '1', '1' ),  -- 1
    ( 'X', '0', '1', '1' )); -- Z

 constant table_or4 :mvl4_table :=
 -- table for or
 --    X    0    1    Z
 -------------------------
   (( 'X', 'X', '1', '1' ),  -- X
    ( 'X', '0', '1', '1' ),  -- 0
    ( '1', '1', '1', '1' ),  -- 1
    ( '1', '1', '1', '1' )); -- Z

 constant table_xor4 :mvl4_table :=
 -- table for xor
 --    X    0    1    Z
 -------------------------
   (( 'X', 'X', 'X', 'X' ),  -- X
    ( 'X', '0', '1', '1' ),  -- 0
    ( 'X', '1', '0', '0' ),  -- 1
    ( 'X', '1', '0', '0' )); -- Z

 -- table for tristate driver, control is column
 constant table_tri4 :mvl4_table :=
 --    X    0    1    Z  <-- control value
 -------------------------
   (( 'X', 'Z', 'X', 'X' ),  -- X
    ( 'X', 'Z', '0', '0' ),  -- 0
    ( 'X', 'Z', '1', '1' ),  -- 1
    ( 'X', 'Z', '1', '1' )); -- Z

 -- table for wired or
 constant table_wiredOr: mvl4_table :=
 --    X    0    1    Z
 -------------------------
   (( 'X', 'X', '1', 'X' ),  -- X
    ( 'X', '0', '1', '0' ),  -- 0
    ( '1', '1', '1', '1' ),  -- 1
    ( 'X', '0', '1', 'Z' )); -- Z

 -- table for wiredX
 constant table_wiredX: mvl4_table :=
 --    X    0    1    Z
 -------------------------
   (( 'X', 'X', 'X', 'X' ),  -- X
    ( 'X', '0', 'X', '0' ),  -- 0
    ( 'X', 'X', '1', '1' ),  -- 1
    ( 'X', '0', '1', 'Z' )); -- Z
 
 function tri_driver(v,c:mvl4) return mvl4;
 function "not"(v: mvl4) return mvl4;  
 function "and"(v1,v2: mvl4) return mvl4;  
 function "or"(v1,v2: mvl4) return mvl4;  
 function "xor"(v1,v2: mvl4) return mvl4;
 function wired_or(V: mvl4_vector) return mvl4;
 function wiredX (V: mvl4_vector) return mvl4;
 function BITtoMVL4 (v: bit) return mvl4;
 function BVtoMVL4V (v: bit_vector) return mvl4_vector;
end system4;



package body system4 is

function "not"(v: mvl4) return mvl4 is
  begin return table_not4(v); end;

function "and"(v1,v2:mvl4) return mvl4 is
   begin return table_and4(v1,v2); end;

function "or"(v1,v2:mvl4) return mvl4 is
   begin return table_or4(v1,v2); end;

function "xor"(v1,v2:mvl4) return mvl4 is
   begin return table_xor4(v1,v2); end;

function tri_driver (v,c : mvl4) return mvl4 is
   begin return table_tri4(v,c); end;

function wiredX(V: mvl4_vector) return mvl4 is
   variable result :mvl4 := 'Z';
   begin
     for i in V'range loop
       result := table_wiredX(result, V(i));
       exit when result = 'X';
     end loop;
     return result;
  end wiredX;

function wired_or(V: mvl4_vector) return mvl4 is
   variable result :mvl4 := 'Z';
   begin
     for i in V'range loop
       result := table_wiredOr(result, V(i));
       exit when result = '1';
     end loop;
     return result;
  end wired_or;

function BITtoMVL4 (v: bit) return mvl4 is
   variable result :mvl4;
   begin
     if v='1' then 
        result := '1';
     else
        result := '0';
     end if;
     return result;
  end;

function BVtoMVL4V (v: bit_vector) return mvl4_vector is
   alias Value: bit_vector (v'length-1 downto 0) is v;
   variable result :mvl4_vector (v'length-1 downto 0);
   begin
     for i in Value'range loop
        if Value(i)='1' then 
           result(i) := '1';
        else
           result(i) := '0';
        end if;
     end loop;
     return result;
  end;

end system4;


use work.system4.all;

------------------------------------------------------------
--  buffer with mvl4
------------------------------------------------------------
entity buffer4 is
    generic (Tmin, Tmax : time);
    port (inval :  in mvl4;
	  outval : inout mvl4);
end buffer4;

architecture beh of buffer4 is
begin
   process(inval)
     begin if (outval /=inval)
	     then  outval <= 'X' after Tmin, inval after Tmax;
	     else  outval <= inval;
	   end if;
    end process;
end beh;

