-- HW11 consists of exercises on the microprogrammed control system 
--described in the book. In this file you find a 'working' version of
-- the partially complete version in the text. There are some changes:
-- CHANGES
-- 1.the registers, busses, and RAM memory all store integers, not bitvectors
-- 2.a resolutin function and resolved integer type is defined
-- 3. the uprogram control memory now contains 13 microcode words
--    rather than 9, as in the textbook.  It appears to work OK. I did
--      try to get the 9 word version going but had difficulty so designed
--      my own microcode. It is very similar to that in the book. The main
--      difference is that I take two control cycles to read memory, one to
--      load the mar, and one to load the mdr. I will discuss this in class.
-- 4. I added an extras control bit to the control word. It is in bit
--    position 9 and it is used to control the 'read' of memory. In the
--    text the 'read' command is the same as the 'mar_in' command and
--    that caused me a problem. So I changed it.
-- 5. Since I did not use bit_vectors on busses I did not use SYSTEM_4
--    or its equivalent
-- 6. I omitted most of the delay characteristics of the hardware
-- 7. I used a single phase system clock and changed control signals
--    on one edge and loaded registers on the other.
-- Otherwise I tried to follow the intentions expressed by the book as
-- much as possible.

-- Here is the HW11 exercise
-- 1. Do a simulation run of the design with some intial values of
--    x and y that you determine. The result of the program is to be
--    (x+y)/2. Does it work for both positive and negative values?
--    Does it work for (x+y) = 0? What value is returned if (x+y) is odd?
--    During the interview you may be asked to conduct a simulation of the
--    design. You should be able to answer questions about how the microcode
--    works. 
-- 2. Is it possible with minor modifications of hardware and/or microcode
--    to improve the design? For example to come closer to the 9 word
--    control memory that I did, using 14. Try it. If you get something
--    good we will give extra credit, more than 100%.
-- 3. The code given in the book forms (x+y)/2 and puts the result in 
--     result. But in the process the value of x is destroyed. Modify
--     the program in the book so that the value of (x+y)/2 is computed
--     and put in result, but the value of x is unchanged.

package intresolve is
  type int_vector is array(integer range <>) of integer;
  function sum (invals :int_vector) return integer;
  subtype resint is sum integer;
end intresolve;

package body intresolve is
  function sum(invals : int_vector) return integer is
    variable temp:integer := 0;
    begin
      for i in invals'range loop
        temp := temp + invals(i);
      end loop;
      return temp;
  end sum;
end intresolve;


use work.intresolve.all;
use work.all;

-- the ram will store integer at integer addresses
-- Notice that each instuction requires three!!! memory locations
entity ram is
   port (  write, read : bit;  -- respond if either is high
          address : integer;
          data : inout resint);
end ram;
 
architecture beh of ram is
-- the intialization corresponds to the prgram of Figure 6.11
-- but the  addresses are changed to reflect the need for three locations
constant x    : integer := 40;
constant temp1: integer := 41;
constant y    : integer := 42;
constant result: integer := 43;
constant temp2: integer := 44;
constant temp3: integer := 45;
constant temp4: integer := 46;
constant two  : integer := 47;
constant one  : integer := 48;
constant mone : integer := 49;
constant pos  : integer := 22;
constant test : integer := 10;
constant neg  : integer := 13;
constant stop : integer     := 0;
constant next3 :integer     := 19;
constant next4 : integer    := 28;

  
signal store : int_vector(0 to 100)  :=
          ( 0,                   --  a branch to this location is 'stop'
            y, temp1,4,          --         temp1 <= temp1 - y
            temp1, x, 7,         --  next1: x <= x +y
            result,result,10,    --  next2: result <= 0
            x,temp2,pos,         --  test : temp2 <= -(x + y); -> pos if neg
            two, temp2,stop,     --  negative: temp2 <=temp2 -2; ->0 if neg
            one, result, next3,  --          result <= result +1;
            two, temp3,neg,      --                    -> negative
            two,x,stop,          -- pos:  x <= x -2;  -> 0 if neg
            mone, result, next4, --       result <= result - 1
            two,temp4,pos,       --       -> pos
            0,0,0,0,0,0,0,0,0,                  -- locations 31..,39
            1,                                  -- x
            0,                                  -- temp1
            1,                                  -- y
            3,                                  -- z alias result
            0,0,0,                              -- temp2,3,4
            2,                                  -- two
            1,                                  -- one
            -1,                                 -- mone
            others => 0);            

begin
     data          <= store(address) after 1 ns when (read = '1') else 0;
    store(address) <= data after 1 ns when (write = '1') else store(address);        
end beh;  
    
use work.intresolve.all; use work.all;
entity urisc is
    port ( data : inout resint := 0;
           address: out integer := 0 ;
           write,read: out  bit);
end urisc;

architecture struc of urisc is


type control_word is array(0 to 9) of bit;
type control_mem  is array(0 to 15) of control_word;

signal ph1 : bit;   -- the main clock
-- registers  and busses
signal r, mdr, bus_a, bus_b, mar : integer := 0;
signal pc : integer := 1;   -- to start in location 1
signal z, nreg, neg : bit;
signal c : integer := 0; --this is the uprogram counter 

-- control points
signal mdr_out, mar_in, n_in, r_in, pc_in, zend, c_in, nnend : bit;
signal mdr_in, comp, pc_out, clear : bit;
signal rd,wrt : bit;   --ram memory signals
signal reg_load, control_load : boolean;

constant mem : control_mem :=
               ( "0100010000",  --0  load MAR <=PC   
                 "0000000001",  --1  load MDR
                 "1100000000",  --2  MAR <= MDR
                 "0000000001",  --3  load MDR
                 "1001000000",  --4  R <= MDR
                 "0100101000",  --5  MAR <+ PC +1, PC <= PC+1
                 "0000000001",  --6  load MDR   
                 "1100000000",  --7  MAR <= MDR
                 "0000000001",  --8  load MDR with second operand
                 "1010001100",  --9  MDR <= MDR - R and write
                 "0100101000",  --10 PC<= PC+1, MAR <= PC +1
                 "0000000001",  --11 MDR <= branch address 
                 "0000101010",  --12 PC loaded from PC + 1 
                 "1000100000",  --13  PC loaded from MDR
       others => "0000000000");

begin


  ph1 <= not ph1 after 10 ns;  -- clock just beats;

  reg_load     <=  ((ph1 = '1') and (ph1 /= ph1'delayed));
  control_load <=  ((ph1 = '0') and (ph1 /= ph1'delayed));


-- synchronous register loads
  PCl: block(reg_load and (pc_in = '1'))
        begin pc    <= guarded bus_b ; end block;
  Rl:  block(reg_load and (r_in = '1'))
        begin r     <= guarded bus_a;  end block;
  MDRl: block(reg_load and ((mdr_in or rd) = '1')) 
        begin
        mdr   <= guarded bus_b   when mdr_in = '1' and reg_load else 
                         data    when (rd = '1') and reg_load
                         else mdr;
       end block;
  MARl: block(reg_load and (mar_in = '1'))  
         begin mar   <= guarded bus_b; end block;
  NREGl: block(reg_load and (n_in = '1'))
         begin nreg <= guarded neg;  end block;
  ZREGl: block(reg_load and (zend = '1'))
         begin z <= guarded '1' when (bus_a = 0) else '0'; end block;
  
-- asychronous events  
  neg   <= '1' when (bus_b < 0) else '0';
  bus_a <= pc     after 1 ns when (pc_out = '1') else mdr ;
  bus_b <= bus_a +1 after 1 ns when (((not comp) and c_in) = '1') else
          bus_a - r after 1 ns when (( comp and c_in) = '1')      else
          bus_a     after 1 ns when (((not comp) and (not c_in)) = '1') else
          bus_a;

-- uprog memory
-- c is the uprog mar  -- c changes on the falling edge
   Cl: block(control_load)
      begin c <= guarded 0 when (clear = '1') else c + 1 ;
      end block;
          
-- asynchronous control signals
   mdr_out <= mem(c)(0);   
   mar_in  <= mem(c)(1);   
   n_in    <= mem(c)(2);   
   r_in    <= mem(c)(3);   
   pc_in   <= mem(c)(4);   
   zend    <= mem(c)(5);   
   c_in    <= mem(c)(6);   
   wrt     <= mem(c)(7);   
   nnend   <= mem(c)(8);   
   rd      <= mem(c)(9); 

   comp    <= n_in; 
   mdr_in  <= n_in;
   clear   <=  '1' when (((z and zend) or (not nreg and nnend)) = '1')
                        or (c = 13)
               else '0';
   pc_out  <= not mdr_out;
-- port connections
   read    <= rd;
   write    <= wrt;
   data    <= mdr when (wrt = '1') else 0;  -- a resolved signal
   address <= mar;

end struc;

   
use work.intresolve.all; use work.all;
entity test_urisc is end test_urisc;

architecture struc of test_urisc is

  component   ram    port (  write, read : bit; 
                            address : integer;
                            data : inout integer); end component;
  component urisc  port ( data : inout resint;
                           address: out integer ;
                           write,read: out  bit); end component;

signal go,wt,rd :bit;
signal addr :integer := 0;
signal dt : resint := 0;
begin
  L3: ram   port map (wt,rd,addr,dt);
  L4: urisc port map (dt,addr,wt,rd);
end struc;


        
